# -*- coding: utf-8 -*-
import os
from tensorflow.python.estimator import export as export_lib

from tensorflow.python.keras._impl.keras.estimator import (
    estimator_lib,
    model_fn_lib,
    ops,
    _clone_and_build_model,
    metrics_module,
    random_seed,
    training_util,
    models,
    saver_lib,
    session,
    logging,
    K
)

from tensorflow.python.saved_model import signature_constants

_DEFAULT_SERVING_KEY = signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY


def _create_keras_model_fn(keras_model, custom_objects=None):
  """Creates model_fn for keras Estimator.

  Args:
    keras_model: an instance of compiled keras model.
    custom_objects: Dictionary for custom objects.

  Returns:
    The model_fn for a keras Estimator.
  """

  def model_fn(features, labels, mode):
    """model_fn for keras Estimator."""
    model = _clone_and_build_model(mode, keras_model, custom_objects, features,
                                   labels)
    # Get inputs to EstimatorSpec
    predictions = dict(zip(model.output_names, model.outputs))

    loss = None
    train_op = None
    eval_metric_ops = None

    # Set loss and metric only during train and evaluate.
    if mode is not model_fn_lib.ModeKeys.PREDICT:
      model._make_train_function()  # pylint: disable=protected-access
      loss = model.total_loss

      if model.metrics:
        eval_metric_ops = {}
        # When each metric maps to an output
        if isinstance(model.metrics, dict):
          for i, output_name in enumerate(model.metrics.keys()):
            metric_name = model.metrics[output_name]
            if callable(metric_name):
              metric_name = metric_name.__name__
            # When some outputs use the same metric
            if list(model.metrics.values()).count(metric_name) > 1:
              metric_name += '_' + output_name
            eval_metric_ops[metric_name] = metrics_module.mean(
                model.metrics_tensors[i - len(model.metrics)])
        else:
          for i, metric_name in enumerate(model.metrics):
            if callable(metric_name):
              metric_name = metric_name.__name__
            eval_metric_ops[metric_name] = metrics_module.mean(
                model.metrics_tensors[i])

    # Set train_op only during train.
    if mode is model_fn_lib.ModeKeys.TRAIN:
      train_op = model.train_function.updates_op

    return model_fn_lib.EstimatorSpec(
        mode=mode,
        predictions=predictions,
        loss=loss,
        train_op=train_op,
        eval_metric_ops=eval_metric_ops,
        export_outputs={
            _DEFAULT_SERVING_KEY: export_lib.export_output.PredictOutput(predictions)
        })

  return model_fn


def _save_first_checkpoint(keras_model, estimator, custom_objects,
                           keras_weights):
  """Save first checkpoint for the keras Estimator.

  Args:
    keras_model: an instance of compiled keras model.
    estimator: keras estimator.
    custom_objects: Dictionary for custom objects.
    keras_weights: A flat list of Numpy arrays for weights of given keras_model.

  Returns:
    The model_fn for a keras Estimator.
  """
  with ops.Graph().as_default() as g, g.device(estimator._device_fn):
    random_seed.set_random_seed(estimator.config.tf_random_seed)
    training_util.create_global_step()
    model = _clone_and_build_model(model_fn_lib.ModeKeys.TRAIN, keras_model,
                                   custom_objects)

    if isinstance(model, models.Sequential):
      model = model.model
    # Load weights and save to checkpoint if there is no checkpoint
    latest_path = saver_lib.latest_checkpoint(estimator.model_dir)
    if not latest_path:
      with session.Session() as sess:
        model.set_weights(keras_weights)
        # Make update ops and initialize all variables.
        if not model.train_function:
          # pylint: disable=protected-access
          model._make_train_function()
          K._initialize_variables(sess)
          # pylint: enable=protected-access
        saver = saver_lib.Saver()
        saver.save(sess, os.path.join(estimator.model_dir, 'keras_model.ckpt'))


def model_to_estimator(keras_model=None,
                       keras_model_path=None,
                       custom_objects=None,
                       model_dir=None,
                       config=None):
  """Constructs an `Estimator` instance from given keras model.

  Args:
    keras_model: Keras model in memory.
    keras_model_path: Directory to a keras model on disk.
    custom_objects: Dictionary for custom objects.
    model_dir: Directory to save Estimator model parameters, graph and etc.
    config: Configuration object.

  Returns:
    An Estimator from given keras model.

  Raises:
    ValueError: if neither keras_model nor keras_model_path was given.
    ValueError: if both keras_model and keras_model_path was given.
    ValueError: if the keras_model_path is a GCS URI.
    ValueError: if keras_model has not been compiled.
  """
  if (not keras_model) and (not keras_model_path):
    raise ValueError(
        'Either keras_model or keras_model_path needs to be provided.')
  if keras_model and keras_model_path:
    raise ValueError(
        'Please specity either keras_model or keras_model_path but not both.')

  if not keras_model:
    if keras_model_path.startswith(
        'gs://') or 'storage.googleapis.com' in keras_model_path:
      raise ValueError(
          '%s is not a local path. Please copy the model locally first.' %
          keras_model_path)
    logging.info('Loading models from %s', keras_model_path)
    keras_model = models.load_model(keras_model_path)
  else:
    logging.info('Using the Keras model from memory.')
    keras_model = keras_model

  if not hasattr(keras_model, 'optimizer'):
    raise ValueError(
        'Given keras model has not been compiled yet. Please compile first '
        'before creating the estimator.')

  keras_weights = keras_model.get_weights()
  keras_model_fn = _create_keras_model_fn(keras_model, custom_objects)
  est = estimator_lib.Estimator(
      keras_model_fn, model_dir=model_dir, config=config)
  # TODO(yifeif): move checkpoint initialization to scaffold.init_fn
  _save_first_checkpoint(keras_model, est, custom_objects, keras_weights)
  return est



# -*- coding: utf-8 -*-
import os
import tempfile
import subprocess
import tensorflow as tf
from tensorflow.core.framework.types_pb2 import DataType
from modelhub.core.models import ModelVersion


def get_meta_graph_def(saved_model_def, tag_set)->tf.MetaGraphDef:
    set_of_tags = set(tag_set.split(','))
    for meta_graph_def in saved_model_def.meta_graphs:
        if set(meta_graph_def.meta_info_def.tags) == set_of_tags:
            return meta_graph_def


def optimize(model_version: ModelVersion, transform_graph_path, output_path=None):
    saved_model_def = model_version.saved_model_def
    for meta_graph_def in saved_model_def.meta_graphs:
        transformed_graph = transform_graph(meta_graph_def, transform_graph_path, model_version.local_path)
        meta_graph_def.graph_def.CopyFrom(transformed_graph)
    if output_path:
        with open(output_path, "wb") as f:
            f.write(saved_model_def.SerializeToString())
    else:
        return saved_model_def


def inspect_meta_graph_io(meta_graph_def):
    output_node_names = [
        output_node.name.split(":")[0]
        for signature_def in meta_graph_def.signature_def.values()
        for output_node in signature_def.outputs.values()
    ]
    input_nodes_name_type_shape = [
        (
            input_node.name.split(":")[0],
            DataType.Name(input_node.dtype)[3:].lower(),
            [dim.size for dim in input_node.tensor_shape.dim]
        )
        for signature_def in meta_graph_def.signature_def.values()
        for input_node in signature_def.inputs.values()
    ]
    return input_nodes_name_type_shape, output_node_names


def freeze_model(meta_graph_def, input_saved_model_dir):
    _, output_node_names = inspect_meta_graph_io(meta_graph_def)
    tags = list(meta_graph_def.meta_info_def.tags)
    from tensorflow.python.tools.freeze_graph import freeze_graph_with_def_protos
    frozen_graph_def = freeze_graph_with_def_protos(
        input_graph_def=meta_graph_def.graph_def,
        input_saver_def=None,
        input_checkpoint=None,
        # input_meta_graph_def=meta_graph_def,
        input_saved_model_dir=input_saved_model_dir,
        saved_model_tags=tags,
        output_node_names=",".join(output_node_names),
        restore_op_name=None,
        filename_tensor_name=None,
        output_graph=None,
        clear_devices=False,
        initializer_nodes=None
    )
    # meta_graph_def.graph_def.CopyFrom(frozen_graph_def)
    return frozen_graph_def


def transform_graph(meta_graph_def, transform_graph_path, input_saved_model_dir):
    with tempfile.NamedTemporaryFile() as f:
        frozen_graph_def = freeze_model(meta_graph_def, input_saved_model_dir)
        input_nodes_name_type_shape, output_node_names = inspect_meta_graph_io(meta_graph_def)

        f.write(frozen_graph_def.SerializeToString())
        f.flush()
        output_name = tempfile.mktemp()
        status = subprocess.run(" ".join([
            transform_graph_path,
            "--in_graph=" + f.name,
            "--out_graph=" + output_name,
            "--inputs='%s'" % ",".join(name for name, type, shape in input_nodes_name_type_shape),
            "--outputs='%s'" % ",".join(output_node_names),
            "--transforms='  %s'" % "\n  ".join(generate_transforms(input_nodes_name_type_shape))
        ]), shell=True)
    if status.returncode:
        raise ValueError("Error transform")

    with open(output_name, "rb") as f:
        new_graph = tf.GraphDef()
        new_graph.ParseFromString(f.read())
    os.unlink(output_name)
    return new_graph


def generate_transforms(input_nodes_name_type_shape):
    return [
        "add_default_attributes",
        # "strip_unused_nodes(type=float, shape="1,299,299,3")",
        generate_strip_unused_nodes(input_nodes_name_type_shape),
        "remove_nodes(op=Identity, op=CheckNumerics)",
        "fold_constants(ignore_errors=true)",
        "fold_batch_norms",
        "fold_old_batch_norms",
        "quantize_weights",
        "quantize_nodes",
        "strip_unused_nodes",
        "sort_by_execution_order",
    ]

    # return [
    #     generate_strip_unused_nodes(input_nodes_name_type_shape),
    #     "fold_constants(ignore_errors=true)",
    #     "fold_batch_norms",
    #     "fold_old_batch_norms",
    #     "quantize_weights",
    # ]


def generate_strip_unused_nodes(input_nodes_name_type_shape):
    args = ", ".join(
        'name={name}, type_for_name={type}, shape_for_name="{shape}"'.format(
            name=name,
            type=type,
            shape=",".join(str(dim) for dim in shape)
        )
        for name, type, shape in input_nodes_name_type_shape
    )
    return "strip_unused_nodes(%s)" % args



# -*- coding: utf-8 -*-
from __future__ import print_function
from modelhub.core.models import Model, ModelVersion
from .base import BaseCommand, argument, option, types, register


@register('checkout')
class Command(BaseCommand):

    arguments = [
        argument("model_name"),
        option("-o", "--output", type=types.Path(), help='the DIR to put checkouted model, by default model will checkout to system application folder'),
        option("-z", "--zip_path", type=types.Path(), help='the DIR to put download model zip, by default zip will be deleted')

    ]

    def run(self, model_name, output=None, zip_path=None):
        """Download a model to local disk."""
        model_name, version = Model.split_name_version(model_name)
        print("verifying...")
        try:
            model = Model.get(name=model_name)
        except Model.DoesNotExist:
            self.echo("Model {model_name} is not exist".format(model_name=model_name))
            return
        try:
            model_version = model.get_version(version)
        except ModelVersion.DoesNotExist:
            self.echo("Model {model.name} has no version: {version}".format(model=model, version=version))
            return

        # if not output:
        #     output = model_version.local_path
        output = model_version.checkout(output, zip_path)
        self.echo("Checkout {model.name}@{model_version.manifest.seq} Done, locate at {output}".format(**locals()))



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tensorflow_serving/apis/input.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tensorflow.core.example import example_pb2 as tensorflow_dot_core_dot_example_dot_example__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tensorflow_serving/apis/input.proto',
  package='tensorflow.serving',
  syntax='proto3',
  serialized_pb=_b('\n#tensorflow_serving/apis/input.proto\x12\x12tensorflow.serving\x1a%tensorflow/core/example/example.proto\"4\n\x0b\x45xampleList\x12%\n\x08\x65xamples\x18\x01 \x03(\x0b\x32\x13.tensorflow.Example\"e\n\x16\x45xampleListWithContext\x12%\n\x08\x65xamples\x18\x01 \x03(\x0b\x32\x13.tensorflow.Example\x12$\n\x07\x63ontext\x18\x02 \x01(\x0b\x32\x13.tensorflow.Example\"\xa1\x01\n\x05Input\x12;\n\x0c\x65xample_list\x18\x01 \x01(\x0b\x32\x1f.tensorflow.serving.ExampleListB\x02(\x01H\x00\x12S\n\x19\x65xample_list_with_context\x18\x02 \x01(\x0b\x32*.tensorflow.serving.ExampleListWithContextB\x02(\x01H\x00\x42\x06\n\x04kindB\x03\xf8\x01\x01\x62\x06proto3')
  ,
  dependencies=[tensorflow_dot_core_dot_example_dot_example__pb2.DESCRIPTOR,])




_EXAMPLELIST = _descriptor.Descriptor(
  name='ExampleList',
  full_name='tensorflow.serving.ExampleList',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='examples', full_name='tensorflow.serving.ExampleList.examples', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=98,
  serialized_end=150,
)


_EXAMPLELISTWITHCONTEXT = _descriptor.Descriptor(
  name='ExampleListWithContext',
  full_name='tensorflow.serving.ExampleListWithContext',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='examples', full_name='tensorflow.serving.ExampleListWithContext.examples', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='context', full_name='tensorflow.serving.ExampleListWithContext.context', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=152,
  serialized_end=253,
)


_INPUT = _descriptor.Descriptor(
  name='Input',
  full_name='tensorflow.serving.Input',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='example_list', full_name='tensorflow.serving.Input.example_list', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('(\001'))),
    _descriptor.FieldDescriptor(
      name='example_list_with_context', full_name='tensorflow.serving.Input.example_list_with_context', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('(\001'))),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='kind', full_name='tensorflow.serving.Input.kind',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=256,
  serialized_end=417,
)

_EXAMPLELIST.fields_by_name['examples'].message_type = tensorflow_dot_core_dot_example_dot_example__pb2._EXAMPLE
_EXAMPLELISTWITHCONTEXT.fields_by_name['examples'].message_type = tensorflow_dot_core_dot_example_dot_example__pb2._EXAMPLE
_EXAMPLELISTWITHCONTEXT.fields_by_name['context'].message_type = tensorflow_dot_core_dot_example_dot_example__pb2._EXAMPLE
_INPUT.fields_by_name['example_list'].message_type = _EXAMPLELIST
_INPUT.fields_by_name['example_list_with_context'].message_type = _EXAMPLELISTWITHCONTEXT
_INPUT.oneofs_by_name['kind'].fields.append(
  _INPUT.fields_by_name['example_list'])
_INPUT.fields_by_name['example_list'].containing_oneof = _INPUT.oneofs_by_name['kind']
_INPUT.oneofs_by_name['kind'].fields.append(
  _INPUT.fields_by_name['example_list_with_context'])
_INPUT.fields_by_name['example_list_with_context'].containing_oneof = _INPUT.oneofs_by_name['kind']
DESCRIPTOR.message_types_by_name['ExampleList'] = _EXAMPLELIST
DESCRIPTOR.message_types_by_name['ExampleListWithContext'] = _EXAMPLELISTWITHCONTEXT
DESCRIPTOR.message_types_by_name['Input'] = _INPUT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ExampleList = _reflection.GeneratedProtocolMessageType('ExampleList', (_message.Message,), dict(
  DESCRIPTOR = _EXAMPLELIST,
  __module__ = 'tensorflow_serving.apis.input_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.serving.ExampleList)
  ))
_sym_db.RegisterMessage(ExampleList)

ExampleListWithContext = _reflection.GeneratedProtocolMessageType('ExampleListWithContext', (_message.Message,), dict(
  DESCRIPTOR = _EXAMPLELISTWITHCONTEXT,
  __module__ = 'tensorflow_serving.apis.input_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.serving.ExampleListWithContext)
  ))
_sym_db.RegisterMessage(ExampleListWithContext)

Input = _reflection.GeneratedProtocolMessageType('Input', (_message.Message,), dict(
  DESCRIPTOR = _INPUT,
  __module__ = 'tensorflow_serving.apis.input_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.serving.Input)
  ))
_sym_db.RegisterMessage(Input)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\370\001\001'))
_INPUT.fields_by_name['example_list'].has_options = True
_INPUT.fields_by_name['example_list']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('(\001'))
_INPUT.fields_by_name['example_list_with_context'].has_options = True
_INPUT.fields_by_name['example_list_with_context']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('(\001'))
try:
  # THESE ELEMENTS WILL BE DEPRECATED.
  # Please use the generated *_pb2_grpc.py files instead.
  import grpc
  from grpc.beta import implementations as beta_implementations
  from grpc.beta import interfaces as beta_interfaces
  from grpc.framework.common import cardinality
  from grpc.framework.interfaces.face import utilities as face_utilities
except ImportError:
  pass
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
from modelhub.core import models
from .base import BaseCommand, argument, option, types, register  # noqa


@register('copy')
class Command(BaseCommand):

    arguments = [
        argument("model_name"),
        argument("new_model_name"),
    ]

    def run(self, model_name, new_model_name):
        """copy model to a new name"""
        model = models.Model.get(model_name)
        model.copy(new_model_name)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tensorflow_serving/sources/storage_path/file_system_storage_path_source.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='tensorflow_serving/sources/storage_path/file_system_storage_path_source.proto',
  package='tensorflow.serving',
  syntax='proto3',
  serialized_pb=_b('\nMtensorflow_serving/sources/storage_path/file_system_storage_path_source.proto\x12\x12tensorflow.serving\"\xd6\x06\n!FileSystemStoragePathSourceConfig\x12Z\n\tservables\x18\x05 \x03(\x0b\x32G.tensorflow.serving.FileSystemStoragePathSourceConfig.ServableToMonitor\x12\x19\n\rservable_name\x18\x01 \x01(\tB\x02\x18\x01\x12\x15\n\tbase_path\x18\x02 \x01(\tB\x02\x18\x01\x12%\n\x1d\x66ile_system_poll_wait_seconds\x18\x03 \x01(\x03\x12(\n fail_if_zero_versions_at_startup\x18\x04 \x01(\x08\x1a\x9d\x03\n\x15ServableVersionPolicy\x12\x64\n\x06latest\x18\x64 \x01(\x0b\x32R.tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicy.LatestH\x00\x12^\n\x03\x61ll\x18\x65 \x01(\x0b\x32O.tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicy.AllH\x00\x12h\n\x08specific\x18\x66 \x01(\x0b\x32T.tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicy.SpecificH\x00\x1a\x1e\n\x06Latest\x12\x14\n\x0cnum_versions\x18\x01 \x01(\r\x1a\x05\n\x03\x41ll\x1a\x1c\n\x08Specific\x12\x10\n\x08versions\x18\x01 \x03(\x03\x42\x0f\n\rpolicy_choice\x1a\xb1\x01\n\x11ServableToMonitor\x12\x15\n\rservable_name\x18\x01 \x01(\t\x12\x11\n\tbase_path\x18\x02 \x01(\t\x12l\n\x17servable_version_policy\x18\x04 \x01(\x0b\x32K.tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicyJ\x04\x08\x03\x10\x04\x62\x06proto3')
)




_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_LATEST = _descriptor.Descriptor(
  name='Latest',
  full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicy.Latest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_versions', full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicy.Latest.num_versions', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=692,
  serialized_end=722,
)

_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_ALL = _descriptor.Descriptor(
  name='All',
  full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicy.All',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=724,
  serialized_end=729,
)

_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_SPECIFIC = _descriptor.Descriptor(
  name='Specific',
  full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicy.Specific',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='versions', full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicy.Specific.versions', index=0,
      number=1, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=731,
  serialized_end=759,
)

_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY = _descriptor.Descriptor(
  name='ServableVersionPolicy',
  full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicy',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='latest', full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicy.latest', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='all', full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicy.all', index=1,
      number=101, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='specific', full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicy.specific', index=2,
      number=102, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_LATEST, _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_ALL, _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_SPECIFIC, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='policy_choice', full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicy.policy_choice',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=363,
  serialized_end=776,
)

_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLETOMONITOR = _descriptor.Descriptor(
  name='ServableToMonitor',
  full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.ServableToMonitor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='servable_name', full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.ServableToMonitor.servable_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='base_path', full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.ServableToMonitor.base_path', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='servable_version_policy', full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.ServableToMonitor.servable_version_policy', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=779,
  serialized_end=956,
)

_FILESYSTEMSTORAGEPATHSOURCECONFIG = _descriptor.Descriptor(
  name='FileSystemStoragePathSourceConfig',
  full_name='tensorflow.serving.FileSystemStoragePathSourceConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='servables', full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.servables', index=0,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='servable_name', full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.servable_name', index=1,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\030\001')), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='base_path', full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.base_path', index=2,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\030\001')), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_system_poll_wait_seconds', full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.file_system_poll_wait_seconds', index=3,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fail_if_zero_versions_at_startup', full_name='tensorflow.serving.FileSystemStoragePathSourceConfig.fail_if_zero_versions_at_startup', index=4,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY, _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLETOMONITOR, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=102,
  serialized_end=956,
)

_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_LATEST.containing_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_ALL.containing_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_SPECIFIC.containing_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['latest'].message_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_LATEST
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['all'].message_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_ALL
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['specific'].message_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_SPECIFIC
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.containing_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.oneofs_by_name['policy_choice'].fields.append(
  _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['latest'])
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['latest'].containing_oneof = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.oneofs_by_name['policy_choice']
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.oneofs_by_name['policy_choice'].fields.append(
  _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['all'])
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['all'].containing_oneof = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.oneofs_by_name['policy_choice']
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.oneofs_by_name['policy_choice'].fields.append(
  _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['specific'])
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['specific'].containing_oneof = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.oneofs_by_name['policy_choice']
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLETOMONITOR.fields_by_name['servable_version_policy'].message_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLETOMONITOR.containing_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG
_FILESYSTEMSTORAGEPATHSOURCECONFIG.fields_by_name['servables'].message_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLETOMONITOR
DESCRIPTOR.message_types_by_name['FileSystemStoragePathSourceConfig'] = _FILESYSTEMSTORAGEPATHSOURCECONFIG
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

FileSystemStoragePathSourceConfig = _reflection.GeneratedProtocolMessageType('FileSystemStoragePathSourceConfig', (_message.Message,), dict(

  ServableVersionPolicy = _reflection.GeneratedProtocolMessageType('ServableVersionPolicy', (_message.Message,), dict(

    Latest = _reflection.GeneratedProtocolMessageType('Latest', (_message.Message,), dict(
      DESCRIPTOR = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_LATEST,
      __module__ = 'tensorflow_serving.sources.storage_path.file_system_storage_path_source_pb2'
      # @@protoc_insertion_point(class_scope:tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicy.Latest)
      ))
    ,

    All = _reflection.GeneratedProtocolMessageType('All', (_message.Message,), dict(
      DESCRIPTOR = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_ALL,
      __module__ = 'tensorflow_serving.sources.storage_path.file_system_storage_path_source_pb2'
      # @@protoc_insertion_point(class_scope:tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicy.All)
      ))
    ,

    Specific = _reflection.GeneratedProtocolMessageType('Specific', (_message.Message,), dict(
      DESCRIPTOR = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_SPECIFIC,
      __module__ = 'tensorflow_serving.sources.storage_path.file_system_storage_path_source_pb2'
      # @@protoc_insertion_point(class_scope:tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicy.Specific)
      ))
    ,
    DESCRIPTOR = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY,
    __module__ = 'tensorflow_serving.sources.storage_path.file_system_storage_path_source_pb2'
    # @@protoc_insertion_point(class_scope:tensorflow.serving.FileSystemStoragePathSourceConfig.ServableVersionPolicy)
    ))
  ,

  ServableToMonitor = _reflection.GeneratedProtocolMessageType('ServableToMonitor', (_message.Message,), dict(
    DESCRIPTOR = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLETOMONITOR,
    __module__ = 'tensorflow_serving.sources.storage_path.file_system_storage_path_source_pb2'
    # @@protoc_insertion_point(class_scope:tensorflow.serving.FileSystemStoragePathSourceConfig.ServableToMonitor)
    ))
  ,
  DESCRIPTOR = _FILESYSTEMSTORAGEPATHSOURCECONFIG,
  __module__ = 'tensorflow_serving.sources.storage_path.file_system_storage_path_source_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.serving.FileSystemStoragePathSourceConfig)
  ))
_sym_db.RegisterMessage(FileSystemStoragePathSourceConfig)
_sym_db.RegisterMessage(FileSystemStoragePathSourceConfig.ServableVersionPolicy)
_sym_db.RegisterMessage(FileSystemStoragePathSourceConfig.ServableVersionPolicy.Latest)
_sym_db.RegisterMessage(FileSystemStoragePathSourceConfig.ServableVersionPolicy.All)
_sym_db.RegisterMessage(FileSystemStoragePathSourceConfig.ServableVersionPolicy.Specific)
_sym_db.RegisterMessage(FileSystemStoragePathSourceConfig.ServableToMonitor)


_FILESYSTEMSTORAGEPATHSOURCECONFIG.fields_by_name['servable_name'].has_options = True
_FILESYSTEMSTORAGEPATHSOURCECONFIG.fields_by_name['servable_name']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\030\001'))
_FILESYSTEMSTORAGEPATHSOURCECONFIG.fields_by_name['base_path'].has_options = True
_FILESYSTEMSTORAGEPATHSOURCECONFIG.fields_by_name['base_path']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\030\001'))
# @@protoc_insertion_point(module_scope)



# encoding=utf8
# author=spenly
# mail=i@spenly.com
from .base import BaseCommand, option, register


class ColorPrint:
    HEADER = '\033[95m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

    @staticmethod
    def show(msg, style):
        print(style + msg + ColorPrint.ENDC)


class GeneralChecker(object):

    def __init__(self, model_name):
        self.model_name = model_name
        self._weights = [100000000, 10000, 1, 0.0001]

    def _get_token_wight(self):
        from modelhub.core.models import Model
        from collections import Counter
        model_names = [model.name for model in Model.all()]
        counter = Counter()
        for mname in model_names:
            for item in zip(mname.split("_")[::-1], self._weights[::-1]):
                counter.update({item[0]: item[1]})
        return counter

    def _generate_new_name(self):
        weights = self._get_token_wight()
        ss = self.model_name.split("_")
        iws = {}
        for item in ss:
            iws[item] = weights.get(item) or 0
        sorted_values = sorted(iws.items(), key=lambda x: x[1], reverse=True)
        return "_".join([item[0] for item in sorted_values])

    def regular_name_check(self):
        new_name = self._generate_new_name()
        if new_name != self.model_name:
            msg = "The model name suggestion: rename [ %s ] to [ %s ]" % (self.model_name, new_name)
            ColorPrint.show(msg, ColorPrint.YELLOW)
        else:
            ColorPrint.show("The model name check passed!", ColorPrint.GREEN)


def benchmark(func, *args, **kwargs):
    import time
    times = 10
    print("Preheating ...")
    items = list(range(times))
    for _ in range(3):
        func(*args, **kwargs)
    print("10 tasks running ...")
    process_time_start = time.clock()
    total_time_start = time.time()
    for _ in items:
        func(*args, **kwargs)
    process_time = (time.clock() - process_time_start) / times
    total_time = (time.time() - total_time_start) / times
    print("tasks running over!")
    ColorPrint.show("CPU / REAL Time Per Task: %f / %f" % (process_time, total_time), ColorPrint.BLUE)


class CodeChecker(object):
    def __init__(self, model_name):
        self.model_name = model_name
        self.model_class = None
        self.break_check = False
        self._interrupted_msg = "checker is interrupted for existed errors or warnings..."

    def prepare_model(self):
        try:
            self.model_class = self._check_import()
            ColorPrint.show("Model instance has been created", ColorPrint.BLUE)
        except Exception as ex:
            ColorPrint.show("Create model instance failed:", ColorPrint.YELLOW)
            ColorPrint.show(str(ex), ColorPrint.RED)
            return False
        return True

    def _check_import(self):
        from importlib import import_module
        ColorPrint.show("trying to import module ...", ColorPrint.BLUE)
        module_fn = "{name}.api_model".format(name=self.model_name)
        app_module = import_module(module_fn)
        return app_module.Model

    def check_properties(self):
        if self.break_check:
            ColorPrint.show(self._interrupted_msg, ColorPrint.YELLOW)
            return
        ColorPrint.show("trying to check properties ...", ColorPrint.BLUE)
        ColorPrint.show("trying to import ...", ColorPrint.BLUE)
        attrs = ["model_name", "INPUTS_SAMPLE", "OUTPUTS_SAMPLE"]
        msg = "The class [%s.api_model.Model] dose not have the property [%s] \n" \
              "For more info about Base model or TFBase model: \n" \
              "[ https://confluence.zhihuiya.com/display/PR/modelhub.framework ]"
        for attr in attrs:
            if not getattr(self.model_class, attr):
                ColorPrint.show(msg % (self.model_name, attr), ColorPrint.RED)
                self.break_check = True
        if not self.break_check:
            ColorPrint.show("properties check passed!", ColorPrint.GREEN)

    def check_sample_running(self):
        from genson import SchemaBuilder
        import json
        from jsonschema import validate
        if self.break_check:
            ColorPrint.show(self._interrupted_msg, ColorPrint.YELLOW)
            return
        ColorPrint.show("trying to check sample ...", ColorPrint.BLUE)
        if self.model_class.TYPE == "NTF":
            model = self.model_class()
        elif self.model_class.TYPE == "TF":
            model = self.model_class(run_mode="local")
        else:
            raise Exception("Invalid model type [%s]" % self.model_class.TYPE)
        inputs_sample = self.model_class.INPUTS_SAMPLE
        inputs_sample_json = json.dumps(inputs_sample)
        outputs_sample = self.model_class.OUTPUTS_SAMPLE
        try:
            outputs = model.run(inputs_sample)
            builder = SchemaBuilder()
            builder.add_object(outputs_sample)
            outputs_schame = builder.to_schema()
            validate(outputs, schema=outputs_schame)
            ColorPrint.show("outsputs schema check passed!", ColorPrint.GREEN)
        except Exception as ex:
            ColorPrint.show("Errors occur when try to run model with INPUTS_SAMPLE:", ColorPrint.RED)
            print(ex)
            print("== INPUTS_SAMPLE ==")
            ColorPrint.show(inputs_sample_json, ColorPrint.BLUE)
            print("===================")
        benchmark(model.run, inputs_sample)

    def check_requirements(self):
        import os
        cmd = "pipreqs --diff requirements.txt --print %s" % self.model_name
        ColorPrint.show("Checking requirements ...", ColorPrint.BLUE)
        os.system(cmd)

    def install(self):
        import os
        import sys
        if not os.path.exists("setup.py"):
            ColorPrint.show("Can not find the file setup.py, please cd project/root/dir ", ColorPrint.YELLOW)
            return False
        ColorPrint.show("going to install [ %s ] in editable mode" % self.model_name, ColorPrint.BLUE)
        # tensorflow-tensorboard<1.6.0,>=1.5.0->tensorflow==1.5.0 required bleach 1.5.0
        os.system("%s -m pip install -U bleach" % sys.executable)
        os.system("%s -m pip install -e ." % sys.executable)
        ColorPrint.show("\n======  Model Inspection Report  ======\n", ColorPrint.GREEN)
        return True


@register("inspect")
class Command(BaseCommand):
    arguments = [
        option("-m", "--model_name", default=""),
    ]

    def run(self, model_name=None):
        """Inspect api_model module before submission, must run at root dir of project (same dir where setup.py locate)"""
        import os
        if not model_name:
            model_name = os.getcwd().split("/")[-1]
            if not os.path.exists(model_name):
                print("Could not determine the model name, please run with option -m")
                return
            ColorPrint.show("Determined model name => %s " % model_name, ColorPrint.YELLOW)
        # model code check
        code_checker = CodeChecker(model_name)
        if code_checker.install() and code_checker.prepare_model():
            # model name check
            name_checker = GeneralChecker(model_name)
            name_checker.regular_name_check()
            code_checker.check_properties()
            code_checker.check_sample_running()
            code_checker.check_requirements()



# -*- coding: utf-8 -*-
import click
from .base import BaseCommand, argument, option, types, register  # noqa
from modelhub.core.models import Model


@register("export")
class Command(BaseCommand):
    arguments = [
        argument("model_name"),
        argument("output_path", type=click.Path()),
        option("-f", "--format", default="GraphDef", type=click.Choice(["GraphDef", "MetaGraphDef", "FreezeGraphDef"])),
        # option("-s", "--signature_def", default="serving_default"),
        option("--tag", default="serve"),
        option("-t", "--text", is_flag=True)
    ]

    def run(self, model_name, output_path, format, tag, text):
        "export tensorflow saved_model with format"
        model_name, version = Model.split_name_version(model_name)
        model = Model.get_local(model_name)
        model_version = model.get_version(version)

        from modelhub.utils.convert_freeze_graph import freeze_model, get_meta_graph_def
        meta_graph_def = get_meta_graph_def(model_version.saved_model_def, tag)

        if format == "MetaGraphDef":
            return self.save_pb(meta_graph_def, output_path, text)

        if format == "FreezeGraphDef":
            return self.save_pb(freeze_model(meta_graph_def, model_version.local_path), output_path, text)
        graph_def = meta_graph_def.graph_def
        if format == "GraphDef":
            return self.save_pb(graph_def, output_path, text)
        raise ValueError("Unsupport format %s" % format)

    def save_pb(self, pb, output_path, text):
        if text:
            with open(output_path, "w") as f:
                f.write(repr(pb))
        else:
            with open(output_path, "wb") as f:
                f.write(pb.SerializeToString())



# -*- coding: utf-8 -*-
from __future__ import print_function
import os
import tempfile
import zipfile
import datetime
import shutil
import json
import pathlib

from google.protobuf import json_format

from modelhub import __version__
from .apis import api
from .utils import cached_property, isidentifier, mkdir
from .import conf

from .protos import model_pb2


class Model(object):

    @classmethod
    def all(cls):
        return [Model(model_name) for model_name in api.ls_subfolder()]

    @classmethod
    def get(cls, name):
        # if name not in api.ls_subfolder():
        #     raise cls.DoesNotExist
        cls.validate_model_name(name)
        model = Model(name)
        try:
            model.manifest
        except api.FileDoesNotExist:
            raise cls.DoesNotExist(name)
        else:
            if model.local_exists:
                model.save(local=True)
        return model

    @classmethod
    def get_local(cls, name, fallback_to_remote=True):
        cls.validate_model_name(name)
        model = Model(name)
        if model.local_exists:
            with open(model.local_manifest_path) as f:
                model.manifest = cls._parse_manifest(json.load(f))
            return model
        if not fallback_to_remote:
            raise cls.DoesNotExist("{name} not exist in local")
        return cls.get(name)

    @classmethod
    def validate_model_name(cls, name):
        if not isidentifier(name):
            raise ValueError("Invalid model_name: {name}. model_name can only contains alphabets, numbers and underscore, and cannot start with number".format(name=name))

    class DoesNotExist(Exception):
        pass

    def __init__(self, name, manifest=None):
        self.name = name
        if manifest is not None:
            self.manifest = manifest

    def __repr__(self):
        return "<Model: %s>" % self.name

    def __eq__(self, other_model):
        if isinstance(other_model, str):
            return self.name == other_model
        return self.name == other_model.name

    manifest_file_name = "manifest.pb.json"

    @property
    def manifest_path(self):
        return os.path.join(self.name, self.manifest_file_name)

    @cached_property
    def manifest(self):
        """
            {
                "name":string,
                "description":string,
                "versions":[
                    {
                        "seq":int,
                        "comment":string,
                        "submit_date":"2012-12-22 11:11:11Z",
                        "file_name":"1.zip",
                        "require_gpu":false,
                    }
                ]
            }
        """
        json_result = api.get_json(self.manifest_path)
        return self._parse_manifest(json_result)

    @classmethod
    def _parse_manifest(cls, dict_):
        pb = model_pb2.Model()
        json_format.ParseDict(dict_, pb)
        return pb

    def save(self, local=False):
        self.manifest.last_update_datetime.FromDatetime(datetime.datetime.now())
        json_value = json_format.MessageToDict(self.manifest)
        if local:
            mkdir(self.local_path)
            with open(self.local_manifest_path, "w") as f:
                json.dump(json_value, f, indent=4)
        else:
            res = api.upload_json(json_value, self.manifest_path, indent=4)
            self.save(local=True)
            return res

    @cached_property
    def versions(self):
        # return api.ls_subfolder(self.name)
        return [ModelVersion(self, version_pb) for version_pb in self.manifest.versions]

    @cached_property
    def latest_version(self):
        try:
            return self.versions[-1]
        except IndexError:
            raise ModelVersion.DoesNotExist("%s has no version exist" % self.name)

    @cached_property
    def latest_local_version(self):

        for version in reversed(self.versions):
            if version.local_exists:
                return version
        raise ModelVersion.DoesNotExist("{name} has no version local exist\nHint: try `modelhub checkout {name}`".format(name=self.name))

    def create_version(self, path, **info):
        return ModelVersion.create(self, path, **info)

    def get_version(self, version=None):
        if not version:
            return self.latest_version
        for model_version in self.versions:
            if model_version.manifest.seq == version:
                return model_version
        raise ModelVersion.DoesNotExist

    @classmethod
    def create(cls, name, description, owner_name, owner_email):
        try:
            cls.get(name)
        except cls.DoesNotExist:
            pass
        else:
            raise ValueError("Model name '{name}' already exist".format(name=name))

        pb = model_pb2.Model(
            name=name,
            description=description,
            owner_name=owner_name,
            owner_email=owner_email,
        )
        pb.create_datetime.FromDatetime(datetime.datetime.now())
        obj = cls(name, manifest=pb)
        obj.save()
        return obj

    def delete(self):
        # TODO: now just delete everything, lately will provide a Trash Can
        return api.delete_file(self.name, recursive=True)

    def delete_local(self):
        for version in self.versions:
            print("deleting version", version.manifest.seq)
            version.local_exists and version.delete_local()
        shutil.rmtree(self.local_path)

    def update(self, **info):
        for k, v in info.values():
            setattr(self.manifest, k, v)
        self.save()

    @property
    def local_path(self):
        return os.path.join(conf.LOCAL_MODEL_DIR, self.name)

    @property
    def local_exists(self):
        return os.path.exists(self.local_manifest_path)

    @property
    def local_manifest_path(self):
        return os.path.join(self.local_path, self.manifest_file_name)

    @staticmethod
    def split_name_version(model_name):
        if "@" in model_name:
            model_name, version = model_name.split("@", 1)
            version = int(version)
        else:
            version = None
        return model_name, version

    def copy(self, new_name, local=False):
        if self.name == new_name:
            return

        cls = type(self)
        manifest = model_pb2.Model()
        manifest.CopyFrom(self.manifest)
        manifest.name = new_name
        new_model = cls(new_name, manifest=manifest)

        if local:
            api.local_link_dir(self.local_path, new_model.local_path)
            return new_model.local_path

        try:
            cls.get(new_name)
        except cls.DoesNotExist:
            pass
        else:
            raise ValueError("Model name '{name}' already exist".format(name=new_name))

        api.copy_dir(self.name, new_model.name)
        self.copy(new_name, local=True)
        new_model.save()


class ModelVersion(object):

    def __init__(self, model, manifest):
        self.model = model
        self.manifest = manifest

    def __repr__(self):
        return "<ModelVersion: %s-%s>" % (self.model.name, self.manifest.seq)

    @property
    def local_path(self):
        return os.path.join(self.model.local_path, str(self.manifest.seq))

    @property
    def saved_model_info(self):
        return self.manifest.saved_model_info[0] if self.manifest.saved_model_info else None

    @property
    def local_exists(self):
        return os.path.exists(self.local_path)

    def download(self, fp):
        return api.get_file(
            fp,
            self.model.name,
            self.manifest.file_name,
            md5=self.manifest.compressed_md5,
            progressbar=True,
            size=self.manifest.compressed_size,
            auto_copy=False
        )

    @classmethod
    def create(cls, model, path, **info):
        # comment is mandatory
        assert "comment" in info
        print("verifying...")
        api.ensure_version("submit")
        versions = model.manifest.versions
        if versions:
            last_version = versions[-1].seq
        else:
            last_version = 0
        version = model_pb2.ModelVersion(
            seq=last_version + 1,
            modelhub_version=__version__,
            **info
        )
        print("parsing...")
        try:
            saved_model_info = cls._parse_saved_model_info(path)
        except cls.ValidationError as e:
            print(e, ", will treate as a NTF model")
            version.is_saved_model = False
            cls.is_valid_ntf_folder(path)
        else:
            version.saved_model_info.extend(saved_model_info)
            version.is_saved_model = True

        file_infos = cls._parse_file_info(path)
        if not file_infos:
            raise cls.ValidationError("No file to submit.")
        version.files.extend(file_infos)

        file_name = "{model.name}-{version.seq}.zip".format(model=model, version=version)
        with tempfile.TemporaryFile() as f:
            print("compressing...")
            cls._compress_model(f, path)
            version.compressed_size = api.get_fp_size(f)
            print("hashing...")
            version.compressed_md5 = api.md5sum(f)
            print("uploading...")
            api.upload_file(f, model.name, file_name, progressbar=True)
        version.file_name = file_name
        version.submit_datetime.FromDatetime(datetime.datetime.now())
        model.manifest.versions.extend([version])
        print("saving...")
        model.save()
        obj = cls(model, version)
        api.local_link_dir(path, obj.local_path)
        return obj

    def delete(self):
        self.model.manifest.versions.remove(self.manifest)
        self.model.save()
        api.delete_file(self.model.name, self.manifest.file_name)
        api.delete_local()

    def delete_local(self):
        assert self.local_exists
        shutil.rmtree(self.local_path)

    def checkout(self, path=None, zip_path=None):
        path = path or self.local_path
        if path == self.local_path and self.local_exists:
            return path
        if zip_path:
            f_fn = lambda: open(zip_path, "wb")
        else:
            f_fn = tempfile.TemporaryFile
        with f_fn() as f:
            print("downloading...")
            self.download(f)
            print("decompressing...")
            self._decompress_model(f, path)
        self.model.save(local=True)
        return path

    @classmethod
    def _parse_saved_model_info(cls, path):
        saved_model_def = cls.is_valid_saved_model(path, load=True)
        if saved_model_def is False:
            raise cls.ValidationError("{path} is not a valid SavedModel path".format(path=path))
        return [
            cls._meta_graph_to_saved_model_info(meta_graph)
            for meta_graph in saved_model_def.meta_graphs
        ]

    @staticmethod
    def _meta_graph_to_saved_model_info(meta_graph):
        saved_model_info = model_pb2.SavedModelInfo(
            tags=meta_graph.meta_info_def.tags,
            tensorflow_version=meta_graph.meta_info_def.tensorflow_version,
            tensorflow_git_version=meta_graph.meta_info_def.tensorflow_git_version,
            # signature_def=meta_graph.signature_def,
            asset_file_def=meta_graph.asset_file_def
        )
        for k, v in meta_graph.signature_def.items():
            saved_model_info.signature_def[k].CopyFrom(v)

        return saved_model_info

    @staticmethod
    def _load_saved_model(path):
        try:
            from tensorflow.core.protobuf.saved_model_pb2 import SavedModel
        except ImportError:
            from tensorflow_protos.core.protobuf.saved_model_pb2 import SavedModel
        saved_model_path = pathlib.Path(path) / "saved_model.pb"
        assert saved_model_path.exists(), "%s is not a valid SavedModel folder" % path
        with saved_model_path.open("rb") as f:
            saved_model = SavedModel()
            saved_model.ParseFromString(f.read())
        return saved_model

    @cached_property
    def saved_model_def(self):
        return self._load_saved_model(self.local_path)

    @classmethod
    def is_valid_saved_model(cls, path, load=False):
        if not os.path.isdir(path):
            return False
        if not os.path.exists(os.path.join(path, "saved_model.pb")):
            return False

        if load:
            try:
                return cls._load_saved_model(path)
            except Exception as e:
                print(e)
                return False
        return True

    @classmethod
    def is_valid_ntf_folder(cls, path):
        if not os.path.isdir(path):
            raise cls.ValidationError("path {path} need to be a folder".format(path=path))
        for item_name in os.listdir(path):
            if not (item_name.startswith(".") or item_name in ("assets", "variables", "saved_model.pb")):
                raise cls.ValidationError("Error: NTF model can only contain 'assets' and 'variables' folder")
            if item_name in ("assets", "variables", "saved_model.pb"):
                if not os.path.isdir(os.path.join(path, item_name)):
                    raise cls.ValidationError("Error: {item_name} must be a folder".format(item_name=item_name))

    @staticmethod
    def _compress_model(f, path):
        with zipfile.ZipFile(f, 'w', zipfile.ZIP_DEFLATED) as zipf:
            for root, dirs, files in os.walk(path):
                for file in files:
                    if file.startswith("."):
                        continue
                    print("compressing", root, file)
                    # dir_name, file_name=os.path.split(path)
                    full_path = os.path.join(root, file)
                    zipf.write(full_path, os.path.relpath(full_path, path))
        f.seek(0)

    @classmethod
    def _parse_file_info(cls, path):
        file_info_cls = model_pb2.FileDef
        file_infos = []
        for root, dirs, files in os.walk(path):
            for file in files:
                if file.startswith("."):
                    continue
                file_path = os.path.join(root, file)
                file_name = os.path.relpath(file_path, path)
                with open(file_path, "rb") as f:
                    md5 = api.md5sum(f)
                file_infos.append(file_info_cls(
                    file_name=file_name,
                    file_size=os.path.getsize(file_path),
                    md5=md5
                ))
        return file_infos

    @staticmethod
    def _decompress_model(f, path):
        with zipfile.ZipFile(f, "r") as zipf:
            zipf.extractall(path)
        f.seek(0)

    class DoesNotExist(Exception):
        pass

    class ValidationError(Exception):
        pass



import appdirs
import os
from .utils import load_conf


DEFAULT_LOCAL_DATA_DIR = appdirs.user_data_dir("modelhub")
LOCAL_DATA_DIR = DEFAULT_LOCAL_DATA_DIR
LOCAL_MODEL_DIR = os.path.join(LOCAL_DATA_DIR, "models")

RC_PATH = "~/.modelhubrc"

USER_NAME = None
USER_EMAIL = None

from modelhub.settings import *   # noqa

RC_PATH = os.path.expanduser(RC_PATH)
if os.path.exists(RC_PATH):
    load_conf(RC_PATH, locals())



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tensorflow_serving/sources/storage_path/static_storage_path_source.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='tensorflow_serving/sources/storage_path/static_storage_path_source.proto',
  package='tensorflow.serving',
  syntax='proto3',
  serialized_pb=_b('\nHtensorflow_serving/sources/storage_path/static_storage_path_source.proto\x12\x12tensorflow.serving\"a\n\x1dStaticStoragePathSourceConfig\x12\x15\n\rservable_name\x18\x01 \x01(\t\x12\x13\n\x0bversion_num\x18\x02 \x01(\x03\x12\x14\n\x0cversion_path\x18\x03 \x01(\tb\x06proto3')
)




_STATICSTORAGEPATHSOURCECONFIG = _descriptor.Descriptor(
  name='StaticStoragePathSourceConfig',
  full_name='tensorflow.serving.StaticStoragePathSourceConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='servable_name', full_name='tensorflow.serving.StaticStoragePathSourceConfig.servable_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version_num', full_name='tensorflow.serving.StaticStoragePathSourceConfig.version_num', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version_path', full_name='tensorflow.serving.StaticStoragePathSourceConfig.version_path', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=96,
  serialized_end=193,
)

DESCRIPTOR.message_types_by_name['StaticStoragePathSourceConfig'] = _STATICSTORAGEPATHSOURCECONFIG
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

StaticStoragePathSourceConfig = _reflection.GeneratedProtocolMessageType('StaticStoragePathSourceConfig', (_message.Message,), dict(
  DESCRIPTOR = _STATICSTORAGEPATHSOURCECONFIG,
  __module__ = 'tensorflow_serving.sources.storage_path.static_storage_path_source_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.serving.StaticStoragePathSourceConfig)
  ))
_sym_db.RegisterMessage(StaticStoragePathSourceConfig)


# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
from setuptools import setup, find_packages

try:
    README = open('README.md').read()
except Exception:
    README = ""
VERSION = "2.0.0-rc1"
TF_REQUIRE_VERSION = "1.4.0"

requirments = [
    "click",
    "boto3",
    "appdirs",
    "grpcio",
    "six",
    "genson",
    "jsonschema",
    "pipreqs-spenly",
    "jinja2"
]

if sys.version_info.major < 3:
    requirments += ["configparser", "pathlib"]

setup(
    name='modelhub',
    version=VERSION,
    description='Modelhub',
    url="http://git.patsnap.com/research/modelhub",
    long_description=README,
    author='Jay Young(yjmade)',
    author_email='yangjian@patsnap.com',
    packages=find_packages(),
    install_requires=requirments,
    extras_require={
        "tf": ["tensorflow>=1.4.0"],
        "tfgpu": ["tensorflow-gpu>=1.4.0"],
        "tf_client": ["redis", "tensorflow>=1.4.0"],
        "tf_serving": ["redis"]
    },
    entry_points={
        'console_scripts': [
            'modelhub=modelhub.commands:main'
        ]
    },
)



# -*- coding: utf-8 -*-
from modelhub.core.models import Model, ModelVersion   # noqa
from modelhub.core import conf
from .base import BaseCommand, argument, option, types, register   # noqa
# import os
import pathlib
from jinja2 import Template


SETUP_PY_TEMPLATE = """
#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
from setuptools import setup, find_packages

try:
    README = open('README.md').read()
except Exception:
    README = ""
VERSION = "0.0.1"

requirments = ["modelhub"{%if tf%}, "tensorflow"{%endif%}]

setup(
    name='{{project_name}}',
    version=VERSION,
    description='{{project_name}}',
    url="http://git.patsnap.com/research/{{project_name}}",
    long_description=README,
    author='{{conf.USER_NAME}}',
    author_email='{{conf.USER_EMAIL}}',
    packages=find_packages(),
    install_requires=requirments,
    extras_require={
        # "extra": ["extra_requirments"],
    },
    entry_points={
        # 'console_scripts': [
        #     'modelhub=modelhub.commands:main'
        # ]
    },
)

"""

GITIGNORE_TEMPLATE = """
# Created by .ignore support plugin (hsz.mobi)

### OSX template
*.DS_Store
.AppleDouble
.LSOverride

# Icon must end with two \r
Icon

# Thumbnails
._*

# Files that might appear in the root of a volume
.DocumentRevisions-V100
.fseventsd
.Spotlight-V100
.TemporaryItems
.Trashes
.VolumeIcon.icns
.com.apple.timemachine.donotpresent

# Directories potentially created on remote AFP share
.AppleDB
.AppleDesktop
Network Trash Folder
Temporary Items
.apdisk

### C++ template
# Compiled Object files
*.slo
*.lo
*.o
*.obj

# Precompiled Headers
*.gch
*.pch

# Compiled Dynamic libraries
*.so
*.dylib
*.dll

# Fortran module files
*.mod
*.smod

# Compiled Static libraries
*.lai
*.la
*.a
*.lib

# Executables
*.exe
*.out
*.app

### IPythonNotebook template
# Temporary data
.ipynb_checkpoints/

### Example user template

# IntelliJ project files
.idea
*.iml
out
gen### Python template
# Byte-compiled / optimized / DLL files
__pycache__/
*.py[cod]
*$py.class

# C extensions

# Distribution / packaging
.Python
env/
build/
develop-eggs/
dist/
downloads/
eggs/
.eggs/
lib/
lib64/
parts/
sdist/
var/
*.egg-info/
.installed.cfg
*.egg

# PyInstaller
#  Usually these files are written by a python script from a template
#  before PyInstaller builds the exe, so as to inject date/other infos into it.
*.manifest
*.spec

# Installer logs
pip-log.txt
pip-delete-this-directory.txt

# Unit test / coverage reports
htmlcov/
.tox/
.coverage
.coverage.*
.cache
nosetests.xml
coverage.xml
*,cover
.hypothesis/

# Translations
*.mo
*.pot

# Django stuff:
*.log
local_settings.py

# Flask stuff:
instance/
.webassets-cache

# Scrapy stuff:
.scrapy

# Sphinx documentation
docs/_build/

# PyBuilder
target/

# IPython Notebook
.ipynb_checkpoints

# pyenv
.python-version

# celery beat schedule file
celerybeat-schedule

# dotenv
.env

# virtualenv
venv/
ENV/

# Spyder project settings
.spyderproject

# Rope project settings
.ropeproject

### JetBrains template
# Covers JetBrains IDEs: IntelliJ, RubyMine, PhpStorm, AppCode, PyCharm, CLion, Android Studio and Webstorm
# Reference: https://intellij-support.jetbrains.com/hc/en-us/articles/206544839

# User-specific stuff:
.idea/workspace.xml
.idea/tasks.xml
.idea/dictionaries
.idea/vcs.xml
.idea/jsLibraryMappings.xml
build-*/

# Sensitive or high-churn files:
.idea/dataSources.ids
.idea/dataSources.xml
.idea/dataSources.local.xml
.idea/sqlDataSources.xml
.idea/dynamic.xml
.idea/uiDesigner.xml

# Gradle:
.idea/gradle.xml
.idea/libraries

# Mongo Explorer plugin:
.idea/mongoSettings.xml

## File-based project format:
*.iws

## Plugin-specific files:

# IntelliJ
/out/

# mpeltonen/sbt-idea plugin
.idea_modules/

# JIRA plugin
atlassian-ide-plugin.xml

# Crashlytics plugin (for Android Studio and IntelliJ)
com_crashlytics_export_strings.xml
crashlytics.properties
crashlytics-build.properties
fabric.properties


*.ipynb
test/samples/

.vscode/
"""

README_MD_TEMPLATE = """
# {{project_name}}

## Introduction
Project owner: {{conf.USER_NAME}} <{{conf.USER_EMAIL}}>

## Installation

```bash
pip install {{project_name}}
```

## Usage

```python
from {{project_name}}.api_model import Model
model=Model()
model.run()
```
"""


INIT_PY_TEMPLATE = """
# -*- coding: utf-8 -*-
import pkg_resources
__version__ = pkg_resources.get_distribution("{{project_name}}").version
"""

API_MODEL_TEMPLATE = """
# -*- coding: utf-8 -*-
from modelhub.framework import BaseModel


class Model(BaseModel):
    model_name = "{{project_name}}"

    INPUTS_SAMPLE = None
    OUTPUTS_SAMPLE = None

    def prepare(self):
        super().prepare()
        # do prepare

    def is_ready(self):
        # check is ready

    def validate_input_data(self, raw_input):
        # do validation
        pass
        return True

    def run_model(self, preprocessed_data):
        # do run
        pass


if __name__=="__main__":
    model = Model()
    output = model.run(model.INPUTS_SAMPLE)
    print(output)

"""

TF_API_MODEL_TEMPLATE = """
# -*- coding: utf-8 -*-
from modelhub.framework import TFBaseModel


class Model(TFBaseModel):
    model_name = "{{project_name}}"
    # local_model_version = 1 # specify this when you need use certain version model

    INPUTS_SAMPLE = None
    OUTPUTS_SAMPLE = None

    def preprocess(raw_input):
        # do preprocess here
        return super().preprocess(raw_input)

    def postprocess(self, tf_output, raw_input, preprocessed_item):
        return super().postprocess(self, tf_output, raw_input, preprocessed_item)


if __name__=="__main__":
    model = Model()
    output = model.run(model.INPUTS_SAMPLE)
    print(output)

"""


@register("putup")
class Command(BaseCommand):
    arguments = [
        argument("project_name"),
        option("-tf", help="Is TensorFlow based model", is_flag=True)
    ]

    def run(self, project_name, tf=False):
        """create new api model project layout"""
        root = pathlib.Path(project_name)
        if root.exists():
            self.echo("Error: {} is existed.".format(root))
            return
        root.mkdir()

        self._create_file(root / "setup.py", SETUP_PY_TEMPLATE, project_name=project_name, conf=conf, tf=tf)
        self._create_file(root / ".gitignore", GITIGNORE_TEMPLATE, project_name=project_name, conf=conf, tf=tf)
        self._create_file(root / "README.md", README_MD_TEMPLATE, project_name=project_name, conf=conf, tf=tf)
        src_dir = root / project_name
        src_dir.mkdir()
        self._create_file(src_dir / "__init__.py", INIT_PY_TEMPLATE, project_name=project_name, conf=conf, tf=tf)
        self._create_file(src_dir / "api_model.py", TF_API_MODEL_TEMPLATE if tf else API_MODEL_TEMPLATE, project_name=project_name, conf=conf, tf=tf)

    def _create_file(self, path, template, **content):
        with path.open("w") as f:
            f.write(Template(template).render(**content).strip())



# -*- coding: utf-8 -*-
from __future__ import print_function
import click
import os
from .base import BaseCommand, argument, option, types, register  # noqa
from modelhub.core.models import Model
from modelhub.core.utils import cached_property

# TODO: move to config
REDIS_URL_ENV_KEY = "RD_REDIS_URL"
REDIS_PREFIX_ENV_KEY = "RD_REDIS_PREFIX"
REDIS_TIMEOUT = 10


@register("runserver")
class Command(BaseCommand):
    arguments = [
        argument("models_name", nargs=-1),
        option("-b", "--bin_path", type=click.Path(exists=True)),
        option("-r", "--register_model", is_flag=True)
    ]

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        self.redis_url = os.getenv(REDIS_URL_ENV_KEY)
        self.redis_prefix = os.getenv(REDIS_PREFIX_ENV_KEY, "")
        # bin_path = "/usr/bin/tensorflow_model_server"

    @cached_property
    def bin_path(self):
        import subprocess
        return subprocess.check_output(["which", "tensorflow_model_server"]).strip().decode()

    def run(self, models_name, bin_path=None, register_model=False):
        """run tensorflow_model_server with models"""
        if not models_name:
            self.echo("no model name provided")
            return

        self.is_register_model = register_model

        if bin_path:
            self.bin_path = bin_path

        models = [Model.get_local(model_name) for model_name in models_name]

        from tempfile import NamedTemporaryFile
        with NamedTemporaryFile("r+") as f:
            f.write(self.generate_model_config(models))
            f.flush()
            self.echo("server bin_path:\t%s" % self.bin_path)
            self.echo("model_config_file:\t%s" % f.name)
            self._run_server(
                models=models,
                model_config_file=f.name,
            )

    def generate_model_config(self, models):
        from tensorflow_serving.config.model_server_config_pb2 import ModelServerConfig, ModelConfigList, ModelConfig

        def build_model_config(model):
            assert model.latest_local_version.manifest.is_saved_model, "%s is not a TensorFlow Model, cannot serving" % model.name
            return ModelConfig(
                name=model.name,
                base_path=model.local_path,
                model_platform="tensorflow"
            )

        res = str(ModelServerConfig(
            model_config_list=ModelConfigList(
                config=[build_model_config(model) for model in models]
            )
        ))
        # print(res)
        return res

    auto_discovery = None

    def _register_model(self, models, port):
        if not self.auto_discovery:
            from modelhub.tf_serving import AutoDiscovery, parse_redis_url, get_local_ip
            self.hostport = "%s:%d" % (get_local_ip(), port)
            if not self.redis_url:
                raise ValueError("No REDIS url configured, should set env '%s'" % REDIS_URL_ENV_KEY)
            ad_args = parse_redis_url(self.redis_url)
            ad_args["redis_prefix"] = self.redis_prefix
            self.auto_discovery = AutoDiscovery(**ad_args)
        for model in models:
            print("register", model.name)
            self.auto_discovery.register(
                model.name,
                model.latest_local_version.manifest.seq,
                self.hostport
            )

    def _run_server(self, models, model_config_file, port=8500, **kwargs):
        from subprocess import Popen, PIPE, TimeoutExpired
        import threading
        popen = Popen(
            [
                self.bin_path,
                "--port=%s" % port,
                "--model_config_file=%s" % model_config_file
            ] + [
                "--%s=%s" % (key, value) for key, value in kwargs.items()
            ],
            stdin=PIPE,
            stderr=PIPE if self.is_register_model else None,
            universal_newlines=True
        )
        try:
            if not self.is_register_model:
                return popen.wait()
            # initialized = False
            while True:
                line = popen.stderr.readline()
                print(line, end="")
                if "Running ModelServer at" in line:
                    # initialized = True
                    self._register_model(models, port)
                    break

            def print_stderr():
                while True:
                    print("start read")
                    line = popen.stderr.readline().strip()
                    if not line:
                        break
                    print(line)
            t = threading.Thread(target=print_stderr, name='print_stderr')
            t.start()

            while True:
                try:
                    popen.wait(timeout=REDIS_TIMEOUT)
                    # print("out", stderr)
                except TimeoutExpired:
                    self._register_model(models, port)
                else:
                    # print(popen.stderr.read().decode())
                    break
        finally:
            print("Killing")
            popen.kill()
